@@ -5,9 +5,12 @@ use super::*;
 pub(crate) enum Keyword {
   Alias,
   AllowDuplicateRecipes,
+  AllowDuplicateVariables,
+  Assert,
   DotenvFilename,
   DotenvLoad,
   DotenvPath,
+  DotenvRequired,
   Else,
   Export,
   Fallback,
@@ -17,12 +20,18 @@ pub(crate) enum Keyword {
   Import,
   Mod,
   PositionalArguments,
+  Quiet,
+  ScriptInterpreter,
   Set,
   Shell,
   Tempdir,
   True,
+  Unexport,
+  Unstable,
   WindowsPowershell,
   WindowsShell,
+  WorkingDirectory,
+  X,
 }
 
 impl Keyword {
@@ -40,3 +49,14 @@ impl<'a> PartialEq<&'a str> for Keyword {
     self.lexeme() == *other
   }
 }
+
+#[cfg(test)]
+mod tests {
+  use super::*;
+
+  #[test]
+  fn keyword_case() {
+    assert_eq!(Keyword::X.lexeme(), "x");
+    assert_eq!(Keyword::IgnoreComments.lexeme(), "ignore-comments");
+  }
+}
