@@ -18,40 +18,43 @@ use {super::*, TokenKind::*};
 /// All methods starting with `parse_*` parse and return a language construct.
 ///
 /// The parser tracks an expected set of tokens as it parses. This set contains
-/// all tokens which would have been accepted at the current point in the parse.
-/// Whenever the parser tests for a token that would be accepted, but does not
-/// find it, it adds that token to the set. When the parser accepts a token, the
-/// set is cleared. If the parser finds a token which is unexpected, the
-/// contents of the set is printed in the resultant error message.
-pub(crate) struct Parser<'tokens, 'src> {
-  /// Source tokens
-  tokens: &'tokens [Token<'src>],
-  /// Index of the next un-parsed token
-  next: usize,
-  /// Current expected tokens
-  expected: BTreeSet<TokenKind>,
-  /// Current recursion depth
-  depth: usize,
-  /// Path to the file being parsed
-  path: PathBuf,
-  /// Depth of submodule being parsed
-  submodule: u32,
+/// all tokens which would have been accepted at the current point in the
+/// parse. Whenever the parser tests for a token that would be accepted, but
+/// does not find it, it adds that token to the set. When the parser accepts a
+/// token, the set is cleared. If the parser finds a token which is unexpected,
+/// the elements of the set are printed in the resultant error message.
+pub(crate) struct Parser<'run, 'src> {
+  expected_tokens: BTreeSet<TokenKind>,
+  file_depth: u32,
+  file_path: &'run Path,
+  import_offsets: Vec<usize>,
+  module_namepath: &'run Namepath<'src>,
+  next_token: usize,
+  recursion_depth: usize,
+  tokens: &'run [Token<'src>],
+  working_directory: &'run Path,
 }
 
-impl<'tokens, 'src> Parser<'tokens, 'src> {
+impl<'run, 'src> Parser<'run, 'src> {
   /// Parse `tokens` into an `Ast`
   pub(crate) fn parse(
-    submodule: u32,
-    path: &Path,
-    tokens: &'tokens [Token<'src>],
+    file_depth: u32,
+    file_path: &'run Path,
+    import_offsets: &[usize],
+    module_namepath: &'run Namepath<'src>,
+    tokens: &'run [Token<'src>],
+    working_directory: &'run Path,
   ) -> CompileResult<'src, Ast<'src>> {
-    Parser {
-      depth: 0,
-      expected: BTreeSet::new(),
-      next: 0,
-      path: path.into(),
-      submodule,
+    Self {
+      expected_tokens: BTreeSet::new(),
+      file_depth,
+      file_path,
+      import_offsets: import_offsets.to_vec(),
+      module_namepath,
+      next_token: 0,
+      recursion_depth: 0,
       tokens,
+      working_directory,
     }
     .parse_ast()
   }
@@ -65,7 +68,7 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   fn unexpected_token(&self) -> CompileResult<'src, CompileError<'src>> {
     self.error(CompileErrorKind::UnexpectedToken {
       expected: self
-        .expected
+        .expected_tokens
         .iter()
         .copied()
         .filter(|kind| *kind != ByteOrderMark)
@@ -81,8 +84,8 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   }
 
   /// An iterator over the remaining significant tokens
-  fn rest(&self) -> impl Iterator<Item = Token<'src>> + 'tokens {
-    self.tokens[self.next..]
+  fn rest(&self) -> impl Iterator<Item = Token<'src>> + 'run {
+    self.tokens[self.next_token..]
       .iter()
       .copied()
       .filter(|token| token.kind != Whitespace)
@@ -107,7 +110,7 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   /// The first token in `kinds` will be added to the expected token set.
   fn next_are(&mut self, kinds: &[TokenKind]) -> bool {
     if let Some(&kind) = kinds.first() {
-      self.expected.insert(kind);
+      self.expected_tokens.insert(kind);
     }
 
     let mut rest = self.rest();
@@ -126,10 +129,10 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
 
   /// Advance past one significant token, clearing the expected token set.
   fn advance(&mut self) -> CompileResult<'src, Token<'src>> {
-    self.expected.clear();
+    self.expected_tokens.clear();
 
-    for skipped in &self.tokens[self.next..] {
-      self.next += 1;
+    for skipped in &self.tokens[self.next_token..] {
+      self.next_token += 1;
 
       if skipped.kind != Whitespace {
         return Ok(*skipped);
@@ -315,6 +318,14 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     self.accept(ByteOrderMark)?;
 
     loop {
+      let mut attributes = self.parse_attributes()?;
+      let mut take_attributes = || {
+        attributes
+          .take()
+          .map(|(_token, attributes)| attributes)
+          .unwrap_or_default()
+      };
+
       let next = self.next()?;
 
       if let Some(comment) = self.accept(Comment)? {
@@ -328,14 +339,26 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
       } else if self.next_is(Identifier) {
         match Keyword::from_lexeme(next.lexeme()) {
           Some(Keyword::Alias) if self.next_are(&[Identifier, Identifier, ColonEquals]) => {
-            items.push(Item::Alias(self.parse_alias(BTreeSet::new())?));
+            items.push(Item::Alias(self.parse_alias(take_attributes())?));
           }
           Some(Keyword::Export) if self.next_are(&[Identifier, Identifier, ColonEquals]) => {
             self.presume_keyword(Keyword::Export)?;
-            items.push(Item::Assignment(self.parse_assignment(true)?));
+            items.push(Item::Assignment(
+              self.parse_assignment(true, take_attributes())?,
+            ));
+          }
+          Some(Keyword::Unexport)
+            if self.next_are(&[Identifier, Identifier, Eof])
+              || self.next_are(&[Identifier, Identifier, Eol]) =>
+          {
+            self.presume_keyword(Keyword::Unexport)?;
+            let name = self.parse_name()?;
+            self.expect_eol()?;
+            items.push(Item::Unexport { name });
           }
           Some(Keyword::Import)
             if self.next_are(&[Identifier, StringToken])
+              || self.next_are(&[Identifier, Identifier, StringToken])
               || self.next_are(&[Identifier, QuestionMark]) =>
           {
             self.presume_keyword(Keyword::Import)?;
@@ -349,25 +372,32 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
             });
           }
           Some(Keyword::Mod)
-            if self.next_are(&[Identifier, Identifier, StringToken])
+            if self.next_are(&[Identifier, Identifier, Comment])
               || self.next_are(&[Identifier, Identifier, Eof])
               || self.next_are(&[Identifier, Identifier, Eol])
+              || self.next_are(&[Identifier, Identifier, Identifier, StringToken])
+              || self.next_are(&[Identifier, Identifier, StringToken])
               || self.next_are(&[Identifier, QuestionMark]) =>
           {
+            let doc = pop_doc_comment(&mut items, eol_since_last_comment);
+
             self.presume_keyword(Keyword::Mod)?;
 
             let optional = self.accepted(QuestionMark)?;
 
             let name = self.parse_name()?;
 
-            let relative = if self.next_is(StringToken) {
+            let relative = if self.next_is(StringToken) || self.next_are(&[Identifier, StringToken])
+            {
               Some(self.parse_string_literal()?)
             } else {
               None
             };
 
             items.push(Item::Module {
+              attributes: take_attributes(),
               absolute: None,
+              doc,
               name,
               optional,
               relative,
@@ -384,13 +414,15 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
           }
           _ => {
             if self.next_are(&[Identifier, ColonEquals]) {
-              items.push(Item::Assignment(self.parse_assignment(false)?));
+              items.push(Item::Assignment(
+                self.parse_assignment(false, take_attributes())?,
+              ));
             } else {
               let doc = pop_doc_comment(&mut items, eol_since_last_comment);
               items.push(Item::Recipe(self.parse_recipe(
                 doc,
                 false,
-                BTreeSet::new(),
+                take_attributes(),
               )?));
             }
           }
@@ -400,34 +432,29 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
         items.push(Item::Recipe(self.parse_recipe(
           doc,
           true,
-          BTreeSet::new(),
+          take_attributes(),
         )?));
-      } else if let Some(attributes) = self.parse_attributes()? {
-        let next_keyword = Keyword::from_lexeme(self.next()?.lexeme());
-        match next_keyword {
-          Some(Keyword::Alias) if self.next_are(&[Identifier, Identifier, ColonEquals]) => {
-            items.push(Item::Alias(self.parse_alias(attributes)?));
-          }
-          _ => {
-            let quiet = self.accepted(At)?;
-            let doc = pop_doc_comment(&mut items, eol_since_last_comment);
-            items.push(Item::Recipe(self.parse_recipe(doc, quiet, attributes)?));
-          }
-        }
       } else {
         return Err(self.unexpected_token()?);
       }
+
+      if let Some((token, attributes)) = attributes {
+        return Err(token.error(CompileErrorKind::ExtraneousAttributes {
+          count: attributes.len(),
+        }));
+      }
     }
 
-    if self.next == self.tokens.len() {
+    if self.next_token == self.tokens.len() {
       Ok(Ast {
-        warnings: Vec::new(),
         items,
+        warnings: Vec::new(),
+        working_directory: self.working_directory.into(),
       })
     } else {
       Err(self.internal_error(format!(
         "Parse completed with {} unparsed tokens",
-        self.tokens.len() - self.next,
+        self.tokens.len() - self.next_token,
       ))?)
     }
   }
@@ -435,7 +462,7 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   /// Parse an alias, e.g `alias name := target`
   fn parse_alias(
     &mut self,
-    attributes: BTreeSet<Attribute>,
+    attributes: BTreeSet<Attribute<'src>>,
   ) -> CompileResult<'src, Alias<'src, Name<'src>>> {
     self.presume_keyword(Keyword::Alias)?;
     let name = self.parse_name()?;
@@ -450,21 +477,27 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   }
 
   /// Parse an assignment, e.g. `foo := bar`
-  fn parse_assignment(&mut self, export: bool) -> CompileResult<'src, Assignment<'src>> {
+  fn parse_assignment(
+    &mut self,
+    export: bool,
+    attributes: BTreeSet<Attribute<'_>>,
+  ) -> CompileResult<'src, Assignment<'src>> {
     let name = self.parse_name()?;
-    self.presume_any(&[Equals, ColonEquals])?;
+    self.presume(ColonEquals)?;
     let value = self.parse_expression()?;
     self.expect_eol()?;
     Ok(Assignment {
+      file_depth: self.file_depth,
       export,
       name,
       value,
+      private: name.lexeme().starts_with('_') || attributes.contains(&Attribute::Private),
     })
   }
 
   /// Parse an expression, e.g. `1 + 2`
   fn parse_expression(&mut self) -> CompileResult<'src, Expression<'src>> {
-    if self.depth == if cfg!(windows) { 48 } else { 256 } {
+    if self.recursion_depth == if cfg!(windows) { 48 } else { 256 } {
       let token = self.next()?;
       return Err(CompileError::new(
         token,
@@ -472,49 +505,38 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
       ));
     }
 
-    self.depth += 1;
+    self.recursion_depth += 1;
 
     let expression = if self.accepted_keyword(Keyword::If)? {
       self.parse_conditional()?
     } else if self.accepted(Slash)? {
       let lhs = None;
-      let rhs = Box::new(self.parse_expression()?);
+      let rhs = self.parse_expression()?.into();
       Expression::Join { lhs, rhs }
     } else {
       let value = self.parse_value()?;
 
       if self.accepted(Slash)? {
         let lhs = Some(Box::new(value));
-        let rhs = Box::new(self.parse_expression()?);
+        let rhs = self.parse_expression()?.into();
         Expression::Join { lhs, rhs }
       } else if self.accepted(Plus)? {
-        let lhs = Box::new(value);
-        let rhs = Box::new(self.parse_expression()?);
+        let lhs = value.into();
+        let rhs = self.parse_expression()?.into();
         Expression::Concatenation { lhs, rhs }
       } else {
         value
       }
     };
 
-    self.depth -= 1;
+    self.recursion_depth -= 1;
 
     Ok(expression)
   }
 
   /// Parse a conditional, e.g. `if a == b { "foo" } else { "bar" }`
   fn parse_conditional(&mut self) -> CompileResult<'src, Expression<'src>> {
-    let lhs = self.parse_expression()?;
-
-    let operator = if self.accepted(BangEquals)? {
-      ConditionalOperator::Inequality
-    } else if self.accepted(EqualsTilde)? {
-      ConditionalOperator::RegexMatch
-    } else {
-      self.expect(EqualsEquals)?;
-      ConditionalOperator::Equality
-    };
-
-    let rhs = self.parse_expression()?;
+    let condition = self.parse_condition()?;
 
     self.expect(BraceL)?;
 
@@ -534,17 +556,51 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     };
 
     Ok(Expression::Conditional {
-      lhs: Box::new(lhs),
-      rhs: Box::new(rhs),
-      then: Box::new(then),
-      otherwise: Box::new(otherwise),
+      condition,
+      then: then.into(),
+      otherwise: otherwise.into(),
+    })
+  }
+
+  fn parse_condition(&mut self) -> CompileResult<'src, Condition<'src>> {
+    let lhs = self.parse_expression()?;
+    let operator = if self.accepted(BangEquals)? {
+      ConditionalOperator::Inequality
+    } else if self.accepted(EqualsTilde)? {
+      ConditionalOperator::RegexMatch
+    } else {
+      self.expect(EqualsEquals)?;
+      ConditionalOperator::Equality
+    };
+    let rhs = self.parse_expression()?;
+    Ok(Condition {
+      lhs: lhs.into(),
+      rhs: rhs.into(),
       operator,
     })
   }
 
+  // Check if the next tokens are a shell-expanded string, i.e., `x"foo"`.
+  //
+  // This function skips initial whitespace tokens, but thereafter is
+  // whitespace-sensitive, so `x"foo"` is a shell-expanded string, whereas `x
+  // "foo"` is not.
+  fn next_is_shell_expanded_string(&self) -> bool {
+    let mut tokens = self
+      .tokens
+      .iter()
+      .skip(self.next_token)
+      .skip_while(|token| token.kind == Whitespace);
+
+    tokens
+      .next()
+      .is_some_and(|token| token.kind == Identifier && token.lexeme() == "x")
+      && tokens.next().is_some_and(|token| token.kind == StringToken)
+  }
+
   /// Parse a value, e.g. `(bar)`
   fn parse_value(&mut self) -> CompileResult<'src, Expression<'src>> {
-    if self.next_is(StringToken) {
+    if self.next_is(StringToken) || self.next_is_shell_expanded_string() {
       Ok(Expression::StringLiteral {
         string_literal: self.parse_string_literal()?,
       })
@@ -563,22 +619,30 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
       if contents.starts_with("#!") {
         return Err(next.error(CompileErrorKind::BacktickShebang));
       }
-
       Ok(Expression::Backtick { contents, token })
     } else if self.next_is(Identifier) {
-      let name = self.parse_name()?;
-
-      if self.next_is(ParenL) {
-        let arguments = self.parse_sequence()?;
-        Ok(Expression::Call {
-          thunk: Thunk::resolve(name, arguments)?,
-        })
+      if self.accepted_keyword(Keyword::Assert)? {
+        self.expect(ParenL)?;
+        let condition = self.parse_condition()?;
+        self.expect(Comma)?;
+        let error = Box::new(self.parse_expression()?);
+        self.expect(ParenR)?;
+        Ok(Expression::Assert { condition, error })
       } else {
-        Ok(Expression::Variable { name })
+        let name = self.parse_name()?;
+
+        if self.next_is(ParenL) {
+          let arguments = self.parse_sequence()?;
+          Ok(Expression::Call {
+            thunk: Thunk::resolve(name, arguments)?,
+          })
+        } else {
+          Ok(Expression::Variable { name })
+        }
       }
     } else if self.next_is(ParenL) {
       self.presume(ParenL)?;
-      let contents = Box::new(self.parse_expression()?);
+      let contents = self.parse_expression()?.into();
       self.expect(ParenR)?;
       Ok(Expression::Group { contents })
     } else {
@@ -590,6 +654,13 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
   fn parse_string_literal_token(
     &mut self,
   ) -> CompileResult<'src, (Token<'src>, StringLiteral<'src>)> {
+    let expand = if self.next_is(Identifier) {
+      self.expect_keyword(Keyword::X)?;
+      true
+    } else {
+      false
+    };
+
     let token = self.expect(StringToken)?;
 
     let kind = StringKind::from_string_or_backtick(token)?;
@@ -634,7 +705,23 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
       unindented
     };
 
-    Ok((token, StringLiteral { kind, raw, cooked }))
+    let cooked = if expand {
+      shellexpand::full(&cooked)
+        .map_err(|err| token.error(CompileErrorKind::ShellExpansion { err }))?
+        .into_owned()
+    } else {
+      cooked
+    };
+
+    Ok((
+      token,
+      StringLiteral {
+        cooked,
+        expand,
+        kind,
+        raw,
+      },
+    ))
   }
 
   /// Parse a string literal, e.g. `"FOO"`
@@ -672,7 +759,7 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     &mut self,
     doc: Option<&'src str>,
     quiet: bool,
-    attributes: BTreeSet<Attribute>,
+    attributes: BTreeSet<Attribute<'src>>,
   ) -> CompileResult<'src, UnresolvedRecipe<'src>> {
     let name = self.parse_name()?;
 
@@ -732,19 +819,32 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
 
     let body = self.parse_body()?;
 
+    let shebang = body.first().map_or(false, Line::is_shebang);
+    let script = attributes
+      .iter()
+      .any(|attribute| matches!(attribute, Attribute::Script(_)));
+
+    if shebang && script {
+      return Err(name.error(CompileErrorKind::ShebangAndScriptAttribute {
+        recipe: name.lexeme(),
+      }));
+    }
+
     Ok(Recipe {
-      shebang: body.first().map_or(false, Line::is_shebang),
+      shebang: shebang || script,
       attributes,
       body,
       dependencies,
       doc,
+      file_depth: self.file_depth,
+      file_path: self.file_path.into(),
+      import_offsets: self.import_offsets.clone(),
       name,
+      namepath: self.module_namepath.join(name),
       parameters: positional.into_iter().chain(variadic).collect(),
-      path: self.path.clone(),
       priors,
       private: name.lexeme().starts_with('_'),
       quiet,
-      depth: self.submodule,
     })
   }
 
@@ -774,13 +874,14 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
 
     if self.accepted(Indent)? {
       while !self.accepted(Dedent)? {
-        let line = if self.accepted(Eol)? {
-          Line {
-            fragments: Vec::new(),
-          }
-        } else {
-          let mut fragments = Vec::new();
+        let mut fragments = Vec::new();
+        let number = self
+          .tokens
+          .get(self.next_token)
+          .map(|token| token.line)
+          .unwrap_or_default();
 
+        if !self.accepted(Eol)? {
           while !(self.accepted(Eol)? || self.next_is(Dedent)) {
             if let Some(token) = self.accept(Text)? {
               fragments.push(Fragment::Text { token });
@@ -793,11 +894,9 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
               return Err(self.unexpected_token()?);
             }
           }
-
-          Line { fragments }
         };
 
-        lines.push(line);
+        lines.push(Line { fragments, number });
       }
     }
 
@@ -845,11 +944,17 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
       Keyword::AllowDuplicateRecipes => {
         Some(Setting::AllowDuplicateRecipes(self.parse_set_bool()?))
       }
+      Keyword::AllowDuplicateVariables => {
+        Some(Setting::AllowDuplicateVariables(self.parse_set_bool()?))
+      }
       Keyword::DotenvLoad => Some(Setting::DotenvLoad(self.parse_set_bool()?)),
+      Keyword::DotenvRequired => Some(Setting::DotenvRequired(self.parse_set_bool()?)),
       Keyword::Export => Some(Setting::Export(self.parse_set_bool()?)),
       Keyword::Fallback => Some(Setting::Fallback(self.parse_set_bool()?)),
       Keyword::IgnoreComments => Some(Setting::IgnoreComments(self.parse_set_bool()?)),
       Keyword::PositionalArguments => Some(Setting::PositionalArguments(self.parse_set_bool()?)),
+      Keyword::Quiet => Some(Setting::Quiet(self.parse_set_bool()?)),
+      Keyword::Unstable => Some(Setting::Unstable(self.parse_set_bool()?)),
       Keyword::WindowsPowershell => Some(Setting::WindowsPowerShell(self.parse_set_bool()?)),
       _ => None,
     };
@@ -861,11 +966,13 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     self.expect(ColonEquals)?;
 
     let set_value = match keyword {
-      Keyword::DotenvFilename => Some(Setting::DotenvFilename(self.parse_string_literal()?.cooked)),
-      Keyword::DotenvPath => Some(Setting::DotenvPath(self.parse_string_literal()?.cooked)),
-      Keyword::Shell => Some(Setting::Shell(self.parse_shell()?)),
-      Keyword::Tempdir => Some(Setting::Tempdir(self.parse_string_literal()?.cooked)),
-      Keyword::WindowsShell => Some(Setting::WindowsShell(self.parse_shell()?)),
+      Keyword::DotenvFilename => Some(Setting::DotenvFilename(self.parse_string_literal()?)),
+      Keyword::DotenvPath => Some(Setting::DotenvPath(self.parse_string_literal()?)),
+      Keyword::ScriptInterpreter => Some(Setting::ScriptInterpreter(self.parse_interpreter()?)),
+      Keyword::Shell => Some(Setting::Shell(self.parse_interpreter()?)),
+      Keyword::Tempdir => Some(Setting::Tempdir(self.parse_string_literal()?)),
+      Keyword::WindowsShell => Some(Setting::WindowsShell(self.parse_interpreter()?)),
+      Keyword::WorkingDirectory => Some(Setting::WorkingDirectory(self.parse_string_literal()?)),
       _ => None,
     };
 
@@ -878,8 +985,8 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     }))
   }
 
-  /// Parse a shell setting value
-  fn parse_shell(&mut self) -> CompileResult<'src, Shell<'src>> {
+  /// Parse interpreter setting value, i.e., `['sh', '-eu']`
+  fn parse_interpreter(&mut self) -> CompileResult<'src, Interpreter<'src>> {
     self.expect(BracketL)?;
 
     let command = self.parse_string_literal()?;
@@ -898,27 +1005,47 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
 
     self.expect(BracketR)?;
 
-    Ok(Shell { arguments, command })
+    Ok(Interpreter { arguments, command })
   }
 
-  /// Parse recipe attributes
-  fn parse_attributes(&mut self) -> CompileResult<'src, Option<BTreeSet<Attribute>>> {
+  /// Item attributes, i.e., `[macos]` or `[confirm: "warning!"]`
+  fn parse_attributes(
+    &mut self,
+  ) -> CompileResult<'src, Option<(Token<'src>, BTreeSet<Attribute<'src>>)>> {
     let mut attributes = BTreeMap::new();
 
-    while self.accepted(BracketL)? {
+    let mut token = None;
+
+    while let Some(bracket) = self.accept(BracketL)? {
+      token.get_or_insert(bracket);
+
       loop {
         let name = self.parse_name()?;
-        let attribute = Attribute::from_name(name).ok_or_else(|| {
-          name.error(CompileErrorKind::UnknownAttribute {
-            attribute: name.lexeme(),
-          })
-        })?;
+
+        let mut arguments = Vec::new();
+
+        if self.accepted(Colon)? {
+          arguments.push(self.parse_string_literal()?);
+        } else if self.accepted(ParenL)? {
+          loop {
+            arguments.push(self.parse_string_literal()?);
+
+            if !self.accepted(Comma)? {
+              break;
+            }
+          }
+          self.expect(ParenR)?;
+        }
+
+        let attribute = Attribute::new(name, arguments)?;
+
         if let Some(line) = attributes.get(&attribute) {
           return Err(name.error(CompileErrorKind::DuplicateAttribute {
             attribute: name.lexeme(),
             first: *line,
           }));
         }
+
         attributes.insert(attribute, name.line);
 
         if !self.accepted(Comma)? {
@@ -932,7 +1059,7 @@ impl<'tokens, 'src> Parser<'tokens, 'src> {
     if attributes.is_empty() {
       Ok(None)
     } else {
-      Ok(Some(attributes.into_keys().collect()))
+      Ok(Some((token.unwrap(), attributes.into_keys().collect())))
     }
   }
 }
@@ -962,7 +1089,15 @@ mod tests {
   fn test(text: &str, want: Tree) {
     let unindented = unindent(text);
     let tokens = Lexer::test_lex(&unindented).expect("lexing failed");
-    let justfile = Parser::parse(0, &PathBuf::new(), &tokens).expect("parsing failed");
+    let justfile = Parser::parse(
+      0,
+      &PathBuf::new(),
+      &[],
+      &Namepath::default(),
+      &tokens,
+      &PathBuf::new(),
+    )
+    .expect("parsing failed");
     let have = justfile.tree();
     if have != want {
       println!("parsed text: {unindented}");
@@ -1000,7 +1135,14 @@ mod tests {
   ) {
     let tokens = Lexer::test_lex(src).expect("Lexing failed in parse test...");
 
-    match Parser::parse(0, &PathBuf::new(), &tokens) {
+    match Parser::parse(
+      0,
+      &PathBuf::new(),
+      &[],
+      &Namepath::default(),
+      &tokens,
+      &PathBuf::new(),
+    ) {
       Ok(_) => panic!("Parsing unexpectedly succeeded"),
       Err(have) => {
         let want = CompileError {
@@ -1013,7 +1155,7 @@ mod tests {
             length,
             path: "justfile".as_ref(),
           },
-          kind: Box::new(kind),
+          kind: kind.into(),
         };
         assert_eq!(have, want);
       }
@@ -1056,6 +1198,18 @@ mod tests {
     tree: (justfile (alias t test)),
   }
 
+  test! {
+    name: single_argument_attribute_shorthand,
+    text: "[group: 'some-group']\nalias t := test",
+    tree: (justfile (alias t test)),
+  }
+
+  test! {
+    name: single_argument_attribute_shorthand_multiple_same_line,
+    text: "[group: 'some-group', group: 'some-other-group']\nalias t := test",
+    tree: (justfile (alias t test)),
+  }
+
   test! {
     name: aliases_multiple,
     text: "alias t := test\nalias b := build",
@@ -1092,6 +1246,15 @@ mod tests {
     tree: (justfile (assignment #export x "hello")),
   }
 
+  test! {
+    name: private_export,
+    text: "
+      [private]
+      export x := 'hello'
+    ",
+    tree: (justfile (assignment #export x "hello")),
+  }
+
   test! {
     name: export_equals,
     text: r#"export x := "hello""#,
@@ -1106,6 +1269,15 @@ mod tests {
     tree: (justfile (assignment x "hello")),
   }
 
+  test! {
+    name: private_assignment,
+    text: "
+      [private]
+      x := 'hello'
+      ",
+    tree: (justfile (assignment x "hello")),
+  }
+
   test! {
     name: assignment_equals,
     text: r#"x := "hello""#,
@@ -1888,6 +2060,12 @@ mod tests {
     tree: (justfile (set allow_duplicate_recipes true)),
   }
 
+  test! {
+    name: set_allow_duplicate_variables_implicit,
+    text: "set allow-duplicate-variables",
+    tree: (justfile (set allow_duplicate_variables true)),
+  }
+
   test! {
     name: set_dotenv_load_true,
     text: "set dotenv-load := true",
@@ -1912,6 +2090,24 @@ mod tests {
     tree: (justfile (set positional_arguments true)),
   }
 
+  test! {
+    name: set_quiet_implicit,
+    text: "set quiet",
+    tree: (justfile (set quiet true)),
+  }
+
+  test! {
+    name: set_quiet_true,
+    text: "set quiet := true",
+    tree: (justfile (set quiet true)),
+  }
+
+  test! {
+    name: set_quiet_false,
+    text: "set quiet := false",
+    tree: (justfile (set quiet false)),
+  }
+
   test! {
     name: set_positional_arguments_false,
     text: "set positional-arguments := false",
@@ -1972,6 +2168,12 @@ mod tests {
     tree: (justfile (set windows_powershell false)),
   }
 
+  test! {
+    name: set_working_directory,
+    text: "set working-directory := 'foo'",
+    tree: (justfile (set working_directory "foo")),
+  }
+
   test! {
     name: conditional,
     text: "a := if b == c { d } else { e }",
@@ -2050,6 +2252,18 @@ mod tests {
     tree: (justfile (mod ? foo "some/file/path.txt")),
   }
 
+  test! {
+    name: assert,
+    text: "a := assert(foo == \"bar\", \"error\")",
+    tree: (justfile (assignment a (assert foo == "bar" "error"))),
+  }
+
+  test! {
+    name: assert_conditional_condition,
+    text: "foo := assert(if a != b { c } else { d } == \"abc\", \"error\")",
+    tree: (justfile (assignment foo (assert (if a != b c d) == "abc" "error"))),
+  }
+
   error! {
     name:   alias_syntax_multiple_rhs,
     input:  "alias foo := bar baz",
@@ -2257,6 +2471,7 @@ mod tests {
     width:  1,
     kind:   UnexpectedToken {
       expected: vec![
+        Identifier,
         StringToken,
       ],
       found: BracketR,
@@ -2299,6 +2514,7 @@ mod tests {
     kind:   UnexpectedToken {
       expected: vec![
         BracketR,
+        Identifier,
         StringToken,
       ],
       found: Eof,
@@ -2405,7 +2621,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "arch",
       found: 1,
-      expected: 0..0,
+      expected: 0..=0,
     },
   }
 
@@ -2419,7 +2635,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "env_var",
       found: 0,
-      expected: 1..1,
+      expected: 1..=1,
     },
   }
 
@@ -2433,7 +2649,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "env",
       found: 3,
-      expected: 1..2,
+      expected: 1..=2,
     },
   }
 
@@ -2447,7 +2663,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "env",
       found: 0,
-      expected: 1..2,
+      expected: 1..=2,
     },
   }
 
@@ -2461,7 +2677,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "env_var_or_default",
       found: 1,
-      expected: 2..2,
+      expected: 2..=2,
     },
   }
 
@@ -2475,7 +2691,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "join",
       found: 1,
-      expected: 2..usize::MAX,
+      expected: 2..=usize::MAX,
     },
   }
 
@@ -2489,7 +2705,7 @@ mod tests {
     kind: FunctionArgumentCountMismatch {
       function: "replace",
       found: 1,
-      expected: 3..3,
+      expected: 3..=3,
     },
   }
 }
